/*
ST7032.cpp

Modified by Kenichi KAKIZAKI
2017/9/24
2015/12/27
2015/12/20
2015/11/03
2015/05/08

Arduino UNO, MEGA, NANO, Leonardo, DUE
Arduino core for ESP8266/ESP32 WiFi chip
*/

// When the display powers up, it is configured as follows:
//
// 1. Display clear
// 2. Function set: 
//    DL = 1; 8-bit interface data 
//    N = 0; 1-line display 
//    F = 0; 5x8 dot character font 
// 3. Display on/off control: 
//    D = 0; Display off 
//    C = 0; Cursor off 
//    B = 0; Blinking off 
// 4. Entry mode set: 
//    I/D = 1; Increment by 1 
//    S = 0; No shift 
//
// Note, however, that resetting the Arduino doesn't reset the LCD, so we
// can't assume that its in that state when a sketch starts (and the
// LiquidCrystal constructor is called).

//#include <stdio.h>
//#include <string.h>
//#include <inttypes.h>
#include <Arduino.h>
#include <Wire.h>

#include <ST7032.h>

#if defined(ARDUINO_ARCH_ESP32)
ST7032::ST7032(uint8_t sda, uint8_t scl)
{
  init(sda, scl);
}

ST7032::ST7032()
{
  init(21, 22);
}

void ST7032::init(uint8_t sda, uint8_t scl)
{
  _addr = ST7032_ADDR ;

  _scl = scl ;
  _sda = sda ;

  _displayfunction = ST7032_CMD_1LINE | ST7032_CMD_5x8DOTS;
  _power = PWR3V3 ;
  _contrast = ST7032_DATA_CONTRAST3V3 ;
}
#elif defined(ARDUINO_ARCH_ESP8266)
ST7032::ST7032(uint8_t sda, uint8_t scl)
{
  init(sda, scl);
}

ST7032::ST7032()
{
  init(4, 5);
}

void ST7032::init(uint8_t sda, uint8_t scl)
{
  _addr = ST7032_ADDR ;

  _scl = scl ;
  _sda = sda ;

  _displayfunction = ST7032_CMD_1LINE | ST7032_CMD_5x8DOTS;
  _power = PWR3V3 ;
  _contrast = ST7032_DATA_CONTRAST3V3 ;
}
#elif defined(ARDUINO_ARCH_SAM)
ST7032::ST7032()
{
  init();
}

void ST7032::init()
{
  _addr = ST7032_ADDR ;

  _displayfunction = ST7032_CMD_1LINE | ST7032_CMD_5x8DOTS;
  _power = PWR3V3 ;
  _contrast = ST7032_DATA_CONTRAST3V3 ;
}
#else
ST7032::ST7032()
{
  init(PWR5V);
}

ST7032::ST7032(uint8_t pwr)
{
  init(pwr);
}

void ST7032::init(uint8_t pwr)
{
  _addr = ST7032_ADDR ;

  _displayfunction = ST7032_CMD_1LINE | ST7032_CMD_5x8DOTS;
  _power = pwr ;
  _contrast = pwr == PWR5V ? ST7032_DATA_CONTRAST5V : ST7032_DATA_CONTRAST3V3 ;
}
#endif


void ST7032::begin(uint8_t cols, uint8_t lines, uint8_t dotsize) {
#if defined(ARDUINO_ARCH_ESP32) || defined(ARDUINO_ARCH_ESP8266)
  Wire.begin(_sda, _scl);
#elif defined(ARDUINO_ARCH_SAM)
  Wire1.begin();
#else
  Wire.begin();
#endif
  if (lines > 1) {
    _displayfunction |= ST7032_CMD_2LINE;
  }
  _numlines = lines;

  setRowOffsets(0x00, 0x40, 0x00 + cols, 0x40 + cols);  

  // for some 1 line displays you can select a 10 pixel high font
  if ((dotsize != ST7032_CMD_5x8DOTS) && (lines == 1)) {
    _displayfunction |= ST7032_CMD_5x10DOTS;
  }

  // SEE PAGE 45/46 FOR INITIALIZATION SPECIFICATION!
  // according to datasheet, we need at least 40ms after power rises above 2.7V
  // before sending commands. Arduino can turn on way before 4.5V so we'll wait 50
  delay(50); 
  
//  uint8_t cmd_init[] = {0x38, 0x39, 0x14, 0x70, 0x56, 0x6c, 0x38, 0x0d, 0x01};

  // first function Set
  command(ST7032_CMD_FUNCTIONSET | _displayfunction);
  // second function Set
  command(ST7032_CMD_FUNCTIONSET | ST7032_CMD_EIS | _displayfunction);
  // Internal OSC frequency
  command(ST7032_CMD_OSCFREQ | 0x04) ;

  // Contrast set
  command(ST7032_CMD_CONTRAST | (((_power == PWR5V) ? ST7032_DATA_CONTRAST5V : ST7032_DATA_CONTRAST3V3) & 0x0f)) ;

  if (_power == PWR5V) {
  // Power/Contrast set
//  command(ST7032_CMD_POWERCONTRAST | ST7032_CMD_BOOSTER | (ST7032_DATA_CONTRAST >> 4)) ;
    command(ST7032_CMD_POWERCONTRAST | (ST7032_DATA_CONTRAST5V >> 4)) ; // booster off
  } else {
  // Power/Contrast set
    command(ST7032_CMD_POWERCONTRAST | ST7032_CMD_BOOSTER | (ST7032_DATA_CONTRAST3V3 >> 4)) ;
  }

  // Follower contril
  command(ST7032_CMD_FOLLOWER | 0x0c) ;

  delay(200); 

  // finally, set # lines, font size, etc.
  command(ST7032_CMD_FUNCTIONSET | _displayfunction);  

  // turn the display on with no cursor or blinking default
  _displaycontrol = ST7032_CMD_DISPLAYON | ST7032_CMD_CURSOROFF | ST7032_CMD_BLINKOFF;  
  display();

  // clear it off
  clear();

  // Initialize to default text direction (for romance languages)
  _displaymode = ST7032_CMD_ENTRYLEFT | ST7032_CMD_ENTRYSHIFTDECREMENT;
  // set the entry mode
  command(ST7032_CMD_ENTRYMODESET | _displaymode);
}

/*
   in some 16x4 LCD when line 3 and 4 are not placed correctly you may try:
     setRowOffsets(0x00, 0x40, 0x14, 0x54)
   or
     setRowOffsets(0x00, 0x40, 0x10, 0x50)
 */
void ST7032::setRowOffsets(int row0, int row1, int row2, int row3)
{
  _row_offsets[0] = row0;
  _row_offsets[1] = row1;
  _row_offsets[2] = row2;
  _row_offsets[3] = row3;
}

/********** high level commands, for the user! */
void ST7032::clear()
{
  command(ST7032_CMD_CLEARDISPLAY);  // clear display, set cursor position to zero
  delayMicroseconds(2000);  // this command takes a long time!
}

void ST7032::home()
{
  command(ST7032_CMD_RETURNHOME);  // set cursor position to zero
  delayMicroseconds(2000);  // this command takes a long time!
}

void ST7032::setCursor(uint8_t col, uint8_t row)
{
  const size_t max_lines = sizeof(_row_offsets) / sizeof(*_row_offsets);
  if ( row >= max_lines ) {
    row = max_lines - 1;    // we count rows starting w/0
  }
  if ( row >= _numlines ) {
    row = _numlines - 1;    // we count rows starting w/0
  }
  
  command(ST7032_CMD_SETDDRAMADDR | (col + _row_offsets[row]));
}

// Turn the display on/off (quickly)
void ST7032::noDisplay() {
  _displaycontrol &= ~ST7032_CMD_DISPLAYON;
  command(ST7032_CMD_DISPLAYCONTROL | _displaycontrol);
}
void ST7032::display() {
  _displaycontrol |= ST7032_CMD_DISPLAYON;
  command(ST7032_CMD_DISPLAYCONTROL | _displaycontrol);
}

// Turns the underline cursor on/off
void ST7032::noCursor() {
  _displaycontrol &= ~ST7032_CMD_CURSORON;
  command(ST7032_CMD_DISPLAYCONTROL | _displaycontrol);
}
void ST7032::cursor() {
  _displaycontrol |= ST7032_CMD_CURSORON;
  command(ST7032_CMD_DISPLAYCONTROL | _displaycontrol);
}

// Turn on and off the blinking cursor
void ST7032::noBlink() {
  _displaycontrol &= ~ST7032_CMD_BLINKON;
  command(ST7032_CMD_DISPLAYCONTROL | _displaycontrol);
}
void ST7032::blink() {
  _displaycontrol |= ST7032_CMD_BLINKON;
  command(ST7032_CMD_DISPLAYCONTROL | _displaycontrol);
}

// These commands scroll the display without changing the RAM
void ST7032::scrollDisplayLeft(void) {
  command(ST7032_CMD_CURSORSHIFT | ST7032_CMD_DISPLAYMOVE | ST7032_CMD_MOVELEFT);
}
void ST7032::scrollDisplayRight(void) {
  command(ST7032_CMD_CURSORSHIFT | ST7032_CMD_DISPLAYMOVE | ST7032_CMD_MOVERIGHT);
}

// This is for text that flows Left to Right
void ST7032::leftToRight(void) {
  _displaymode |= ST7032_CMD_ENTRYLEFT;
  command(ST7032_CMD_ENTRYMODESET | _displaymode);
}

// This is for text that flows Right to Left
void ST7032::rightToLeft(void) {
  _displaymode &= ~ST7032_CMD_ENTRYLEFT;
  command(ST7032_CMD_ENTRYMODESET | _displaymode);
}

// This will 'right justify' text from the cursor
void ST7032::autoscroll(void) {
  _displaymode |= ST7032_CMD_ENTRYSHIFTINCREMENT;
  command(ST7032_CMD_ENTRYMODESET | _displaymode);
}

// This will 'left justify' text from the cursor
void ST7032::noAutoscroll(void) {
  _displaymode &= ~ST7032_CMD_ENTRYSHIFTINCREMENT;
  command(ST7032_CMD_ENTRYMODESET | _displaymode);
}

// set display contrast
void ST7032::setContrast(uint8_t value) {
  _contrast = value & 0x3f ;

  command(ST7032_CMD_FUNCTIONSET | ST7032_CMD_EIS | _displayfunction);

  // Contrast set
  command(ST7032_CMD_CONTRAST | (_contrast & 0x0f)) ;
  // Power/Contrast set
  if (_power == PWR5V) {
    command(ST7032_CMD_POWERCONTRAST | (_contrast >> 4)) ; // booster off
  } else {
    command(ST7032_CMD_POWERCONTRAST | ST7032_CMD_BOOSTER | (_contrast >> 4)) ;
  }

  command(ST7032_CMD_FUNCTIONSET | _displayfunction);
}

// set display contrast
uint8_t ST7032::getContrast() {
  return _contrast ;
}

// Allows us to fill the first 8 CGRAM locations
// with custom characters
void ST7032::createChar(uint8_t location, uint8_t charmap[]) {
  location &= 0x7; // we only have 8 locations 0-7
  command(ST7032_CMD_SETCGRAMADDR | (location << 3));
  for (int i=0; i<8; i++) {
    write(charmap[i]);
  }
  home() ;
}

/*********** mid level commands, for sending data/cmds */

inline void ST7032::command(uint8_t value) {
  send(value, LOW);
}

inline size_t ST7032::write(uint8_t value) {
  send(value, HIGH);
  return 1; // assume sucess
}

/************ low level data pushing commands **********/

// write either command or data, with automatic 4/8-bit selection
void ST7032::send(uint8_t value, uint8_t mode) {
#if defined(ARDUINO_ARCH_SAM)
  Wire1.beginTransmission(_addr);
  Wire1.write(mode ? 0x40: 0x00);
  Wire1.write(value);
  Wire1.endTransmission();
#else
  Wire.beginTransmission(_addr);
  Wire.write(mode ? 0x40: 0x00);
  Wire.write(value);
  Wire.endTransmission();
#endif
  delayMicroseconds(25);
}
